library(tree)

# ----1.----

data <- read.csv2("bank-full.csv")
data <- subset(data, select = -duration)

data <- data.frame(lapply(data, function(x)
  if (is.character(x))
    factor(x)
  else
    x))

n = dim(data)[1]
set.seed(12345)
id = sample(1:n, floor(n * 0.4))
train = data[id, ]
id1 = setdiff(1:n, id)
set.seed(12345)
id2 = sample(id1, floor(n * 0.3))
valid = data[id2, ]
id3 = setdiff(id1, id2)
test = data[id3, ]

# ----2.----

# Misclassification rate
misclass = function(X, X_fit) {
  n = length(X)
  return(1 - sum(diag(table(X, X_fit))) / n)
}

# Default settings

fit2a <- tree(y ~ ., data = train)

summary(fit2a)

pred2a <- predict(fit2a, valid, "class")

misclass(valid$y, pred2a)

# Smallest allowed node size equal to 7000

fit2b <- tree(y ~ ., data = train, control = tree.control(nrow(train), minsize = 7000))

summary(fit2b)

pred2b <- predict(fit2b, valid, "class")

misclass(valid$y, pred2b)

# Minimum deviance to 0.0005

fit2c <- tree(y ~ .,
              data = train,
              control = tree.control(nrow(train), mindev = 0.0005))

summary(fit2c)

pred2c <- predict(fit2c, valid, "class")

misclass(valid$y, pred2c)

# ----3.----

train_score <- rep(0, 50)
valid_score <- rep(0, 50)

for (i in 2:50) {
  pruned_tree = prune.tree(fit2c, best = i)
  pred <- predict(pruned_tree, newdata = valid, type = "tree")
  train_score[i] <- deviance(pruned_tree)
  valid_score[i] <- deviance(pred)
}

y_range <- range(c(valid_score[2:50], train_score[2:50]))

plot(
  2:50,
  train_score[2:50],
  type = "b",
  col = "red",
  ylim = c(min(y_range), max(y_range)),
  log = "y",
  main = "Deviance depending on number of leaves",
  xlab = "Number of leaves",
  ylab = "Log of deviance",
  pch = 16,
)
points(2:50, valid_score[2:50], type = "b", col = "blue")
grid()
legend(
  "topright",
  legend = c("Training", "Validation"),
  fill = c("red", "blue"),
  pch = c(16, 1)
)

optimal_leaves <- which.min(valid_score[-1])
optimal_tree <- prune.tree(fit2c, best = optimal_leaves)

summary(optimal_tree)

# ----4.----

TP <- function(x, x_pred) {
  return(table(x, x_pred)[2, 2])
}

TN <- function(x, x_pred) {
  return(table(x, x_pred)[1, 1])
}

FP <- function(x, x_pred) {
  return(table(x, x_pred)[1, 2])
}

FN <- function(x, x_pred) {
  return(table(x, x_pred)[2, 1])
}

P <- function(x, x_pred) {
  return(TP(x, x_pred) + FN(x, x_pred))
}

N <- function(x, x_pred) {
  return(TN(x, x_pred) + FP(x, x_pred))
}

TPR <- function(x, x_pred) {
  return(TP(x, x_pred) / P(x, x_pred))
}

FPR <- function(x, x_pred) {
  return(FP(x, x_pred) / N(x, x_pred))
}

precision <- function(x, x_pred) {
  return(TP(x, x_pred) / (TP(x, x_pred) + FP(x, x_pred)))
}

recall <- function(x, x_pred) {
  return(TP(x, x_pred) / (TP(x, x_pred) + FN(x, x_pred)))
}

accuracy <- function(x, x_pred) {
  return((TP(x, x_pred) + TN(x, x_pred)) / (P(x, x_pred) + N(x, x_pred)))
}

F1 <- function(x, x_pred) {
  return(2 * precision(x, x_pred) * recall(x, x_pred) / (precision(x, x_pred) + recall(x, x_pred)))
}

y_hat4 <- predict(optimal_tree, newdata = test, type = "class")

table(test$y, y_hat4)

accuracy(test$y, y_hat4)

F1(test$y, y_hat4)

# ----5.----

fit5 <- tree(y ~ ., data = train)

probs5 <- predict(fit5, newdata = test, type = "vector")

y_hat5 <- ifelse(probs5[, "no"] / probs5[, "yes"] > 5, "no", "yes")

table(test$y, y_hat5)

accuracy(test$y, y_hat5)

F1(test$y, y_hat5)

# ----6.----

logistic_model <- glm(y ~ ., family = binomial, data = train)

pi <- seq(0.05, 0.95, by = 0.05)

tree_tpr <- numeric(length(pi))
tree_fpr <- numeric(length(pi))

tree_precision <- numeric(length(pi))
tree_recall <- numeric(length(pi))

logistic_tpr <- numeric(length(pi))
logistic_fpr <- numeric(length(pi))

logistic_precision <- numeric(length(pi))
logistic_recall <- numeric(length(pi))

for (i in 1:length(pi)) {
  logistic_pred <- predict(logistic_model, test, type = "response")
  logistic_class <- ifelse(logistic_pred > pi[i], 1, 0)
  logistic_class <- factor(logistic_class,
                           levels = c(0, 1),
                           labels = c("no", "yes"))
  
  logistic_tpr[i] <- TPR(test$y, logistic_class)
  logistic_fpr[i] <- FPR(test$y, logistic_class)
  logistic_precision[i] <- precision(test$y, logistic_class)
  logistic_recall[i] <- recall(test$y, logistic_class)
  
  tree_pred <- predict(optimal_tree, test, type = "vector")[, 2]
  tree_class <- ifelse(tree_pred > pi[i], 1, 0)
  tree_class <- factor(tree_class,
                       levels = c(0, 1),
                       labels = c("no", "yes"))
  
  tree_tpr[i] <- TPR(test$y, tree_class)
  tree_fpr[i] <- FPR(test$y, tree_class)
  tree_precision[i] <- precision(test$y, tree_class)
  tree_recall[i] <- recall(test$y, tree_class)
  
}

plot(
  logistic_fpr,
  logistic_tpr,
  type = "b",
  ylim = c(0, 1),
  xlim = c(0, 1),
  col = "red",
  main = "ROC",
  xlab = "FPR",
  ylab = "TPR",
  pch = 16,
)
points(tree_fpr, tree_tpr, col = "blue", type = "b")
legend(
  "bottomright",
  legend = c("Logistic regression model", "Optimal tree"),
  fill = c("red", "blue"),
  pch = c(16, 1)
)

plot(
  logistic_recall,
  logistic_precision,
  type = "b",
  ylim = c(0, 1),
  xlim = c(0, 1),
  col = "red",
  main = "Precision depending on recall",
  xlab = "Recall",
  ylab = "Precision",
  pch = 16,
)
points(tree_recall,
       tree_precision,
       col = "blue",
       type = "b")
legend(
  "bottomleft",
  legend = c("Logistic regression model", "Optimal tree"),
  fill = c("red", "blue"),
  pch = c(16, 1)
)